"""Data utils."""
import logging
import math
import random
from collections import defaultdict

import embeddix

logger = logging.getLogger(__name__)

__all__ = ('load_splits_dict')


def _load_splits_dict(left_idx, right_idx, sim, kfold_size):
    kfold_dict = defaultdict(defaultdict)
    len_test_set = max(math.floor(len(sim)*kfold_size), 1)
    fold = 1
    max_num_fold = math.floor(len(sim) / len_test_set)
    while fold <= max_num_fold:
        test_start_idx = (fold-1)*len_test_set
        test_end_len = test_start_idx + len_test_set
        test_end_idx = test_end_len if test_end_len <= len(sim) else len(sim)
        test_split_idx_set = set(range(test_start_idx, test_end_idx))
        kfold_dict[fold]['test'] = {
            'left_idx': left_idx[test_start_idx:test_end_idx],
            'right_idx': right_idx[test_start_idx:test_end_idx],
            'sim': sim[test_start_idx:test_end_idx]
        }
        train_split_idx_set = set(range(0, len(sim))) - test_split_idx_set
        train_split_idx_list = sorted(list(train_split_idx_set))
        kfold_dict[fold]['train'] = {
            'left_idx': [left_idx[idx] for idx in train_split_idx_list],
            'right_idx': [right_idx[idx] for idx in train_split_idx_list],
            'sim': [sim[idx] for idx in train_split_idx_list]
        }
        fold += 1
    return kfold_dict


def load_splits_dict(dataset, vocab, kfold_size):
    """Return a kfold train/test dict.

    The dict has the form dict[kfold_num] = {train_dict, test_dict} where
    train/test_dict = {left_idx, right_idx, sim}
    the train/test division follows kfold_size expressed as a precentage
    of the total dataset dedicated to testing.
    kfold_size should be a float > 0 and <= 0.5
    """
    left_idx, right_idx, f_sim = embeddix.load_dataset(dataset, vocab)
    if kfold_size == 0:
        return {
            1: {
                'train': {
                    'left_idx': left_idx,
                    'right_idx': right_idx,
                    'sim': f_sim
                },
                'test': {
                    'left_idx': left_idx,
                    'right_idx': right_idx,
                    'sim': f_sim
                }
            }
        }
    # shuffle everything before converting to kfold
    shuffled_zip = list(zip(left_idx, right_idx, f_sim))
    random.shuffle(shuffled_zip)
    unz_shuffle = list(zip(*shuffled_zip))
    left_idx = list(unz_shuffle[0])
    right_idx = list(unz_shuffle[1])
    f_sim = list(unz_shuffle[2])
    return _load_splits_dict(left_idx, right_idx, f_sim, kfold_size)



"""Evaluate a given Numpy distributional model against the MEN dataset."""
import logging

import embeddix

__all__ = ('evaluate', 'is_improving', 'is_degrading')

logger = logging.getLogger(__name__)


def is_degrading(metric_value, best_metric_value, metric):
    """Return true if metric value is degrading."""
    if metric not in ['spr', 'rmse', 'pearson', 'both']:
        raise Exception('Unsupported metric: {}'.format(metric))
    if metric == 'both':
        spr = metric_value[0]
        rmse = metric_value[1]
        best_spr = best_metric_value[0]
        best_rmse = best_metric_value[1]
        return spr < best_spr or rmse > best_rmse
    if metric in ['spr', 'pearson']:
        return metric_value < best_metric_value
    return metric_value > best_metric_value


def is_improving(metric_value, best_metric_value, metric):
    """Return true if metric value improving."""
    if metric not in ['spr', 'rmse', 'pearson', 'both']:
        raise Exception('Unsupported metric: {}'.format(metric))
    if metric == 'both':
        spr = metric_value[0]
        rmse = metric_value[1]
        best_spr = best_metric_value[0]
        best_rmse = best_metric_value[1]
        return spr > best_spr and rmse < best_rmse
    if metric in ['spr', 'pearson']:
        return metric_value > best_metric_value
    # for rmse we want to lower the loss
    return metric_value < best_metric_value


def evaluate(model, splits, metric):
    """Evaluate model on dataset splits with metric."""
    if metric not in ['spr', 'rmse', 'pearson', 'both']:
        raise Exception('Unsupported metric: {}'.format(metric))
    left_vectors = model[splits['left_idx']]
    right_vectors = model[splits['right_idx']]
    model_sim = embeddix.similarity(left_vectors, right_vectors)
    if metric == 'spr':
        return embeddix.spearman(splits['sim'], model_sim)
    if metric == 'pearson':
        return embeddix.pearson(splits['sim'], model_sim)
    if metric == 'rmse':
        return embeddix.rmse(splits['sim'], model_sim)
    return (embeddix.spearman(splits['sim'], model_sim),
            embeddix.rmse(splits['sim'], model_sim))



"""Welcome to entropix.

This is the entry point of the application.
"""
import os

import argparse
import logging
import logging.config

import embeddix

import entropix.utils.config as cutils
import entropix.utils.data as dutils
import entropix.core.sampler as sampler
import entropix.core.evaluator as evaluator

logging.config.dictConfig(
    cutils.load(
        os.path.join(os.path.dirname(__file__), 'logging', 'logging.yml')))

logger = logging.getLogger(__name__)


# pylint: disable=R0914
def _sample(model_filepath, vocab_filepath, dataset, kfold_size, mode, metric,
            shuffle, max_num_threads, limit):
    if mode not in ['seq', 'limit']:
        raise Exception('Invalid sampling mode: {}'.format(mode))
    model = embeddix.load_dense(model_filepath)
    vocab = embeddix.load_vocab(vocab_filepath)
    splits_dict = dutils.load_splits_dict(dataset, vocab, kfold_size)
    if mode == 'seq':
        logger.info('Sampling dimensions in seq mode over {} dims, '
                    'optimizing on {} using {}'
                    .format(model.shape[1], dataset, metric))
        sampled_dims = sampler.sample_seq(model, splits_dict, kfold_size,
                                          metric, shuffle, max_num_threads)
    if mode == 'limit':
        if kfold_size > 0:
            raise Exception('kfolding currently unsupported in limit mode')
        sampled_dims = sampler.sample_limit(model, splits_dict[1]['train'],
                                            metric, limit)
    results = {}
    for fold, dims in sampled_dims.items():
        results[fold] = {
            'dims': dims,
            'train': evaluator.evaluate(
                model[:, dims], splits_dict[fold]['train'], metric=metric),
            'test': evaluator.evaluate(
                model[:, dims], splits_dict[fold]['test'], metric=metric)
        }
    return results


def sample(args):
    """Sample dimensions from model."""
    results = _sample(args.model, args.vocab, args.dataset, args.kfold_size,
                      args.mode, args.metric, args.shuffle, args.num_threads,
                      args.limit)
    if args.dump:
        if args.kfold_size != 0:
            raise Exception('Cannot dump dims for kfolds')
        output_fpath = '{}.sampled.mode-{}.dataset-{}.metric-{}.dims'.format(
            os.path.abspath(args.model).split('.npy')[0], args.mode,
            args.dataset, args.metric)
        with open(output_fpath, 'w', encoding='utf-8') as output_str:
            for dim in sorted(results[1]['dims']):
                print(dim, file=output_str)


# pylint: disable=C0103
def restricted_kfold_size(x):
    """Restrict kfold-size values."""
    x = float(x)
    if x < 0.0 or x > 0.5:
        raise argparse.ArgumentTypeError('{} kfold-size not in range [0, 0.5]'
                                         .format(x))
    return x


def main():
    """Launch entropix."""
    parser = argparse.ArgumentParser(prog='entropix')
    subparsers = parser.add_subparsers()
    parser_sample = subparsers.add_parser(
        'sample', formatter_class=argparse.RawTextHelpFormatter,
        help='find best dim config that maximize dataset score')
    parser_sample.set_defaults(func=sample)
    parser_sample.add_argument('-m', '--model', required=True,
                               help='absolute path to .singvectors.npy file')
    parser_sample.add_argument('-v', '--vocab', required=True,
                               help='asbolute path to .vocab file')
    parser_sample.add_argument('-d', '--dataset', required=True,
                               choices=['men', 'simlex', 'simverb'],
                               help='dataset to optimize on')
    parser_sample.add_argument('-s', '--shuffle', action='store_true',
                               help='whether or not to shuffle at each iter')
    parser_sample.add_argument('-z', '--mode', choices=['seq', 'limit'],
                               help='which version of the algorithm to use')
    parser_sample.add_argument('-l', '--limit', type=int, default=0,
                               help='max number of dim in limit mode')
    parser_sample.add_argument('-x', '--kfold-size',
                               type=restricted_kfold_size, default=0,
                               help='determine size of kfold. Should be in '
                                    '[0, 0.5], that is, less than 50% of '
                                    'total dataset size')
    parser_sample.add_argument('-c', '--metric', required=True,
                               choices=['spr', 'rmse', 'pearson', 'both'],
                               help='which eval metric to use. '
                                    'both is spr+rmse')
    parser_sample.add_argument('-n', '--num-threads', type=int, default=1,
                               help='number of threads to use for parallel '
                                    'processing of kfold validation')
    parser_sample.add_argument('--dump', action='store_true',
                               help='if set, will save sampled dims to file')
    args = parser.parse_args()
    args.func(args)



"""Dimensionality reduction through dimensionality selection."""
import logging
import random
import functools
import multiprocessing

import numpy as np

import entropix.core.evaluator as evaluator

__all__ = ('sample_seq', 'sample_limit')

logger = logging.getLogger(__name__)


def _init_eval_metric(metric):
    if metric not in ['spr', 'rmse', 'pearson', 'both']:
        raise Exception('Unsupported metric: {}'.format(metric))
    if metric in ['spr', 'pearson']:
        return 0
    if metric == 'rmse':
        return 10**15.
    return (0, 10**15)


# pylint: disable=C0103
def sample_limit(model, train_splits, metric, limit):
    """Sample dimensions in limit mode."""
    best_metric = _init_eval_metric(metric)
    dims = []
    max_num_dim_best = 0
    alldims = list(range(model.shape[1]))
    for k in range(limit):
        best_dim_idx = -1
        least_worst_dim = -1
        least_worst_metric = _init_eval_metric(metric)
        for dim_idx in alldims:
            if dim_idx in dims:
                continue
            dims.append(dim_idx)
            eval_metric = evaluator.evaluate(
                model[:, dims], train_splits, metric=metric)
            if evaluator.is_improving(eval_metric,
                                      best_metric,
                                      metric=metric):
                best_metric = eval_metric
                best_dim_idx = dim_idx
            elif evaluator.is_improving(eval_metric,
                                        least_worst_metric,
                                        metric=metric):
                least_worst_metric = eval_metric
                least_worst_dim = dim_idx
            dims.pop()
        if best_dim_idx == -1:
            logger.info('Could not find a better metric with {} '
                        'dims. Added least worst dim to continue'.format(k+1))
            dims.append(least_worst_dim)
        else:
            dims.append(best_dim_idx)
            max_num_dim_best = len(dims)
            logger.info('Current best {} = {} with dims = {}'
                        .format(metric, best_metric, dims))
    logger.info('Best eval metrix = {} with ndims = {}'
                .format(best_metric, max_num_dim_best))
    return {1: dims}  # to remain consistent with sample_seq return


def sample_seq_reduce(model, splits_dict, dims, step, fold, metric,
                      best_train_eval_metric):
    """Remove dimensions that do not negatively impact scores on train."""
    logger.info('Reducing dimensions while maintaining highest score '
                'on eval metric {}. Step = {}. Best train eval metric = {}'
                .format(metric, step, best_train_eval_metric))
    dims_set = set(dims)
    for dim_idx in dims:
        dims_set.remove(dim_idx)
        train_eval_metric = evaluator.evaluate(
            model[:, list(dims_set)], splits_dict[fold]['train'],
            metric=metric)
        if evaluator.is_degrading(train_eval_metric,
                                  best_train_eval_metric,
                                  metric=metric):
            dims_set.add(dim_idx)
            continue
        logger.info('Constant best train {} = {} for fold {} removing '
                    'dim_idx = {}. New ndim = {}'
                    .format(metric, train_eval_metric, fold,
                            dim_idx, len(dims_set)))
        best_train_eval_metric = train_eval_metric
    logger.info('Finished reducing dims')
    keep = list(sorted(dims_set, key=dims.index))
    if len(keep) != len(dims):
        step += 1
        keep, best_train_eval_metric = sample_seq_reduce(
            model, splits_dict, keep, step, fold, metric,
            best_train_eval_metric)
    return keep, best_train_eval_metric


def sample_seq_add(model, splits_dict, keep, alldims, metric, fold,
                   best_train_eval_metric):
    """Add dimensions that improve scores on train."""
    logger.info('Increasing dimensions to maximize score on eval metric '
                '{}. Best train eval metric = {}'
                .format(metric, best_train_eval_metric))
    dims = [idx for idx in alldims if idx not in keep]
    added_counter = 0
    for idx, dim_idx in enumerate(dims):
        keep.append(dim_idx)
        train_eval_metric = evaluator.evaluate(
            model[:, keep], splits_dict[fold]['train'], metric=metric)
        if evaluator.is_improving(train_eval_metric,
                                  best_train_eval_metric,
                                  metric=metric):
            added_counter += 1
            best_train_eval_metric = train_eval_metric
            logger.info('New best train {} = {} on fold {} with ndim = {} '
                        'at idx = {} and dim_idx = {}'.format(
                            metric, best_train_eval_metric, fold,
                            len(keep), idx, dim_idx))
        else:
            keep.pop()
    return keep, best_train_eval_metric


def _sample_seq(model, splits_dict, keep, alldims, metric, fold):
    best_train_eval_metric = evaluator.evaluate(
        model[:, keep], splits_dict[fold]['train'], metric=metric)
    logger.debug('Initial train eval metric = {}'.format(
        best_train_eval_metric))
    keep, best_train_eval_metric = sample_seq_add(
        model, splits_dict, keep, alldims, metric, fold,
        best_train_eval_metric)
    keep, best_train_eval_metric = sample_seq_reduce(
        model, splits_dict, keep, 1, fold, metric, best_train_eval_metric)
    return fold, keep


def sample_seq(model, splits_dict, kfold_size, metric, shuffle,
               max_num_threads):
    """Sample dimensions in sequential mode."""
    alldims = list(range(model.shape[1]))
    if shuffle:
        random.shuffle(alldims)
        keep = np.random.choice(list(range(model.shape[1])),
                                size=2, replace=False).tolist()
    else:
        keep = [0, 1]  # select the first two
    # sample dimensons multi-threaded on all kfolds
    num_folds = len(splits_dict.keys())
    logger.info('Applying kfolding with k={} folds where '
                'each test fold is of size {} and accounts for '
                '{}% of the data'
                .format(num_folds, len(splits_dict[1]['test']['sim']),
                        kfold_size*100))
    num_threads = num_folds if num_folds <= max_num_threads \
        else max_num_threads
    with multiprocessing.Pool(num_threads) as pool:
        _sample = functools.partial(_sample_seq, model, splits_dict, keep,
                                    alldims, metric)
        sampled_dims = {}
        for fold, keep in pool.imap_unordered(_sample,
                                              range(1, num_folds+1)):
            sampled_dims[fold] = keep
    return sampled_dims
